# adapted from https://github.com/ray-project/ray/blob/5ec63ccc5f351f143bd325e764d1f2a523023ca9/rllib/examples/models/action_mask_model.py

from gym.spaces import Dict, Box
from ray.rllib.models.torch.fcnet import FullyConnectedNetwork as TorchFC
from ray.rllib.models.torch.torch_modelv2 import TorchModelV2
from ray.rllib.utils.framework import try_import_tf, try_import_torch
from ray.rllib.utils.torch_utils import FLOAT_MIN
import numpy as np

tf1, tf, tfv = try_import_tf()
torch, nn = try_import_torch()


class TorchActionMaskModel(TorchModelV2, nn.Module):
    """PyTorch version
    Model that handles simple discrete action masking.
    This assumes the outputs are logits for a single Categorical action dist.
    Getting this to work with a more complex output (e.g., if the action space
    is a tuple of several distributions) is also possible but left as an
    exercise to the reader.
    """

    def __init__(
            self,
            obs_space,
            action_space,
            num_outputs,
            model_config,
            name,
            **kwargs,
    ):
        orig_space = getattr(obs_space, "original_space", obs_space)
        assert (
                isinstance(orig_space, Dict)
                and "action_mask" in orig_space.spaces
                and "observations" in orig_space.spaces
        )

        TorchModelV2.__init__(
            self, obs_space, action_space, num_outputs, model_config, name, **kwargs
        )
        nn.Module.__init__(self)

        self.internal_model = TorchFC(
            orig_space["observations"],
            action_space,
            num_outputs,
            model_config,
            name + "_internal",
        )

        # disable action masking --> will likely lead to invalid actions
        self.no_masking = False
        if "no_masking" in model_config["custom_model_config"]:
            self.no_masking = model_config["custom_model_config"]["no_masking"]

    def forward(self, input_dict, state, seq_lens):
        # Extract the available actions tensor from the observation.
        action_mask = input_dict["obs"]["action_mask"]

        # Compute the unmasked logits.
        logits, _ = self.internal_model({"obs": input_dict["obs"]["observations"]})

        # If action masking is disabled, directly return unmasked logits
        if self.no_masking:
            return logits, state

        # Convert action_mask into a [0.0 || -inf]-type mask.
        inf_mask = torch.clamp(torch.log(action_mask), min=FLOAT_MIN)
        masked_logits = logits + inf_mask

        # Return masked logits.
        return masked_logits, state

    def value_function(self):
        return self.internal_model.value_function()


class TorchPlayerRelation(TorchModelV2, nn.Module):
    """PyTorch version
    Model that handles simple discrete action masking.
    This assumes the outputs are logits for a single Categorical action dist.
    Getting this to work with a more complex output (e.g., if the action space
    is a tuple of several distributions) is also possible but left as an
    exercise to the reader.
    """

    def __init__(
            self,
            obs_space,
            action_space,
            num_outputs,
            model_config,
            name,
            hidden_dim=16,
            statistics_lenght=19,  # card sizes
            card_size=[4, 3],  # card sizes
            **kwargs,
    ):
        orig_space = getattr(obs_space, "original_space", obs_space)
        assert (
                isinstance(orig_space, Dict)
                and "action_mask" in orig_space.spaces
                and "observations" in orig_space.spaces
        )
        self.statistics_lenght = statistics_lenght
        self.card_size = card_size

        TorchModelV2.__init__(
            self, obs_space, action_space, num_outputs, model_config, name, **kwargs
        )
        nn.Module.__init__(self)

        self.card3_conv = nn.Sequential(
            nn.Linear(card_size[1], hidden_dim),
            nn.ReLU(),
        )

        self.rows_conv = nn.Sequential(
            nn.Linear(card_size[0] * hidden_dim, hidden_dim * 2),
            nn.ReLU(),
            nn.Linear(hidden_dim * 2, hidden_dim * 2),
            nn.ReLU(),
        )

        self.fc_player_pairs = nn.Sequential(
            nn.Linear(hidden_dim * 4 + 2, hidden_dim * 4),
            nn.ReLU(),
            nn.Linear(hidden_dim * 4, hidden_dim * 4),
            nn.ReLU(),
        )
        self.f_norm = nn.LayerNorm(hidden_dim * 4)

        self.internal_fcmodel = TorchFC(
            Box(low=-10, high=127, shape=(hidden_dim * 4 + statistics_lenght,)),
            action_space,
            num_outputs,
            model_config,
            name + "_internal_out",
        )

        # disable action masking --> will likely lead to invalid actions
        self.no_masking = False
        if "no_masking" in model_config["custom_model_config"]:
            self.no_masking = model_config["custom_model_config"]["no_masking"]

    def factorize_observation(self, obs):
        batch_size, obs_size = obs.size()
        flat_statistics = obs[:, :self.statistics_lenght]

        player_cards = obs[:, self.statistics_lenght:].reshape(batch_size, -1, self.card_size[0], self.card_size[1])

        return flat_statistics, player_cards

    @staticmethod
    def _add_entitiy_indices_as_feature(input_v):
        """add range over top of embedding
        Args:
            input_v (_type_): _description_
        Returns:
            _type_: _description_
        """
        orig_shape = input_v.shape

        features_indices = np.arange(
            orig_shape[1],  # add features for indices
        )
        std = features_indices.std()
        if std:
            features_indices = (features_indices - features_indices.mean()) / std

        features_indices = np.tile(features_indices, (orig_shape[0], 1)).reshape(orig_shape[0], orig_shape[1], -1)
        return torch.cat(
            (
                torch.from_numpy(features_indices).to(dtype=input_v.dtype, device=input_v.device),
                input_v
            ),
            dim=2
        )

    def internal_relation(self, flat_statistics, player_cards):
        """
            flat_statistics: shape (batch, self.statistics_lenght,)
            player_cards: shape (batch, n_players, card_size[0]=4, card_size[1]=3)

        return:
            game_embedding shape (batch, self.statistics_lenght + self.hidden_size * 4)
        """
        shape_cards = player_cards.size()
        # player_cards_over_rows shape = (batch, n_players, card_size[0] * hidden_dim)
        player_cards_over_rows = self.card3_conv(player_cards).flatten(2)

        # player_cards_over_rows shape = (batch, n_players, 2 * hidden_dim)
        player_encoding = self.rows_conv(player_cards_over_rows)

        player_encoding = self._add_entitiy_indices_as_feature(player_encoding)

        n_players = player_encoding.size(1)
        xi = player_encoding.repeat(1, n_players, 1)
        xj = player_encoding.unsqueeze(2)
        xj = xj.repeat(1, 1, n_players, 1).view(player_encoding.size(0), -1, player_encoding.size(2))
        # pair_concat.size=(batch_size, n_players*n_players, 4 * hidden_dim + 2)
        #
        pair_concat = torch.cat((xi, xj), dim=2)
        player_relations = self.fc_player_pairs(pair_concat)

        embedding_player = self.f_norm(torch.sum(player_relations, dim=1))

        game_embedding = torch.cat((flat_statistics.flatten(1), embedding_player.flatten(1)), dim=1)

        # re
        return game_embedding

    def forward(self, input_dict, state, seq_lens):
        # Extract the available actions tensor from the observation.
        action_mask = input_dict["obs"]["action_mask"]
        observation = input_dict["obs"]["observations"]
        unsqueezed = False
        if observation.ndim == 1:
            unsqueezed = True
            observation = torch.unsqueeze(observation, 0)

        flat_statistics, player_cards = self.factorize_observation(observation)
        # Compute the unmasked logits.
        game_embedding = self.internal_relation(flat_statistics, player_cards)
        logits, _ = self.internal_fcmodel({"obs": game_embedding})
        # If action masking is disabled, directly return unmasked logits
        if unsqueezed:
            logits = torch.squeeze(logits)

        if self.no_masking:
            return logits, state

        # Convert action_mask into a [0.0 || -inf]-type mask.
        inf_mask = torch.clamp(torch.log(action_mask), min=FLOAT_MIN)
        masked_logits = logits + inf_mask

        # Return masked logits.
        return masked_logits, state

    def value_function(self):
        return self.internal_fcmodel.value_function()


# Extending tianshou collector class to work with manual policy (for user input)
import time
import warnings
from typing import Any, Callable, Dict, List, Optional, Union

import gym
import numpy as np
import torch

from tianshou.data import (
    Batch,
    ReplayBuffer,
)
from tianshou.env import BaseVectorEnv, DummyVectorEnv
from tianshou.policy import BasePolicy

from tianshou.data.collector import Collector


class ManualPolicyCollector(Collector):
    def __init__(
        self,
        policy: BasePolicy,
        env: Union[gym.Env, BaseVectorEnv],
        buffer: Optional[ReplayBuffer] = None,
        preprocess_fn: Optional[Callable[..., Batch]] = None,
        exploration_noise: bool = False,
    ) -> None:
        super(ManualPolicyCollector, self).__init__(policy=policy, env=env, exploration_noise=exploration_noise)

    # Custom function to collect the result of an inputted action
    def collect_result(
            self,
            action: int = None,
            render: Optional[float] = None,
            no_grad: bool = True,
            gym_reset_kwargs: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """Collect the results of an inputted action..

        :param int action: the action you want to collect the results for.
        :param float render: the sleep time between rendering consecutive frames.
            Default to None (no rendering).
        :param bool no_grad: whether to retain gradient in policy.forward(). Default to
            True (no gradient retaining).
        :param gym_reset_kwargs: extra keyword arguments to pass into the environment's
            reset function. Defaults to None (extra keyword arguments)

        :return: A dict including the following keys

            * ``n/ep`` collected number of episodes.
            * ``n/st`` collected number of steps.
            * ``rews`` array of episode reward over collected episodes.
            * ``lens`` array of episode length over collected episodes.
            * ``idxs`` array of episode start index in buffer over collected episodes.
            * ``rew`` mean of episodic rewards.
            * ``len`` mean of episodic lengths.
            * ``rew_std`` standard error of episodic rewards.
            * ``len_std`` standard error of episodic lengths.
        """
        assert not self.env.is_async, "Please use AsyncCollector if using async venv."

        ready_env_ids = np.arange(self.env_num)

        start_time = time.time()

        step_count = 0
        episode_count = 0
        episode_rews = []
        episode_lens = []
        episode_start_indices = []

        while True:
            assert len(self.data) == len(ready_env_ids)
            # restore the state: if the last state is None, it won't store
            last_state = self.data.policy.pop("hidden_state", None)

            # use hard coded action (rather than using a policy or randomly sampling)
            self.data.update(act=action)

            # hard code to only update for a single step
            n_step = 1

            # get bounded and remapped actions first (not saved into buffer)
            action_remap = self.policy.map_action(self.data.act)
            # step in env
            result = self.env.step(action_remap, ready_env_ids)  # type: ignore
            if len(result) == 5:
                obs_next, rew, terminated, truncated, info = result
                done = np.logical_or(terminated, truncated)
            elif len(result) == 4:
                obs_next, rew, done, info = result
                if isinstance(info, dict):
                    truncated = info["TimeLimit.truncated"]
                else:
                    truncated = np.array(
                        [
                            info_item.get("TimeLimit.truncated", False)
                            for info_item in info
                        ]
                    )
                terminated = np.logical_and(done, ~truncated)
            else:
                raise ValueError()

            self.data.update(
                obs_next=obs_next,
                rew=rew,
                terminated=terminated,
                truncated=truncated,
                done=done,
                info=info
            )
            if self.preprocess_fn:
                self.data.update(
                    self.preprocess_fn(
                        obs_next=self.data.obs_next,
                        rew=self.data.rew,
                        done=self.data.done,
                        info=self.data.info,
                        policy=self.data.policy,
                        env_id=ready_env_ids,
                    )
                )

            if render:
                self.env.render()
                if render > 0 and not np.isclose(render, 0):
                    time.sleep(render)

            # add data into the buffer
            ptr, ep_rew, ep_len, ep_idx = self.buffer.add(
                self.data, buffer_ids=ready_env_ids
            )

            # collect statistics
            step_count += len(ready_env_ids)

            if np.any(done):
                env_ind_local = np.where(done)[0]
                env_ind_global = ready_env_ids[env_ind_local]
                episode_count += len(env_ind_local)
                episode_lens.append(ep_len[env_ind_local])
                episode_rews.append(ep_rew[env_ind_local])
                episode_start_indices.append(ep_idx[env_ind_local])
                # now we copy obs_next to obs, but since there might be
                # finished episodes, we have to reset finished envs first.
                self._reset_env_with_ids(
                    env_ind_local, env_ind_global, gym_reset_kwargs
                )
                for i in env_ind_local:
                    self._reset_state(i)

                # remove surplus env id from ready_env_ids
                # to avoid bias in selecting environments


            self.data.obs = self.data.obs_next

            if (n_step and step_count >= n_step):
                break

        # generate statistics
        self.collect_step += step_count
        self.collect_episode += episode_count
        self.collect_time += max(time.time() - start_time, 1e-9)

        if episode_count > 0:
            rews, lens, idxs = list(
                map(
                    np.concatenate,
                    [episode_rews, episode_lens, episode_start_indices]
                )
            )
            rew_mean, rew_std = rews.mean(), rews.std()
            len_mean, len_std = lens.mean(), lens.std()
        else:
            rews, lens, idxs = np.array([]), np.array([], int), np.array([], int)
            rew_mean = rew_std = len_mean = len_std = 0

        return {
            "n/ep": episode_count,
            "n/st": step_count,
            "rews": rews,
            "lens": lens,
            "idxs": idxs,
            "rew": rew_mean,
            "len": len_mean,
            "rew_std": rew_std,
            "len_std": len_std,
        }


# adapted from https://github.com/Farama-Foundation/PettingZoo/blob/master/tutorials/Tianshou/3_cli_and_logging.py
"""
This is a full example of using Tianshou with MARL to train agents, complete with argument parsing (CLI) and logging.

Author: Will (https://github.com/WillDudley)

Python version used: 3.8.10

Requirements:
pettingzoo == 1.22.0
git+https://github.com/thu-ml/tianshou
"""

import argparse
import os
from copy import deepcopy
from typing import Optional, Tuple

import gym
import numpy as np
import torch
from tianshou.data import Collector, VectorReplayBuffer
from tianshou.env import DummyVectorEnv
from tianshou.env.pettingzoo_env import PettingZooEnv
from tianshou.policy import BasePolicy, DQNPolicy, MultiAgentPolicyManager, RandomPolicy
from tianshou.trainer import offpolicy_trainer
from tianshou.utils import TensorboardLogger
from tianshou.utils.net.common import Net
from torch.utils.tensorboard import SummaryWriter

from gobblet import gobblet_v1
from gobblet.game.collector_manual_policy import ManualPolicyCollector
from gobblet.game.utils import GIFRecorder
from gobblet.game.greedy_policy import GreedyPolicy
import time


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument("--seed", type=int, default=1626)
    parser.add_argument("--eps-test", type=float, default=0.05)
    parser.add_argument("--eps-train", type=float, default=0.1)
    parser.add_argument("--buffer-size", type=int, default=20000)
    parser.add_argument("--lr", type=float, default=1e-4) # TODO: Changing this to 1e-5 for some reason makes it pause after 3 or 4 epochs
    parser.add_argument(
        "--gamma", type=float, default=0.9, help="a smaller gamma favors earlier win"
    )
    parser.add_argument("--n-step", type=int, default=3)
    parser.add_argument("--target-update-freq", type=int, default=320)
    parser.add_argument("--epoch", type=int, default=50)
    parser.add_argument("--step-per-epoch", type=int, default=1000)
    parser.add_argument("--step-per-collect", type=int, default=10)
    parser.add_argument("--update-per-step", type=float, default=0.1)
    parser.add_argument("--batch-size", type=int, default=64)
    parser.add_argument(
        "--hidden-sizes", type=int, nargs="*", default=[128, 128, 128, 128]
    )
    parser.add_argument("--training-num", type=int, default=10)
    parser.add_argument("--test-num", type=int, default=10)
    parser.add_argument("--logdir", type=str, default="log")
    parser.add_argument("--render", type=float, default=0.1)
    parser.add_argument("--render_mode", type=str, default="human", choices=["human","rgb_array", "text", "text_full"], help="Choose the rendering mode for the game.")
    parser.add_argument("--debug", action="store_true", help="Flag to enable to print extra debugging info")
    parser.add_argument("--self_play", action="store_true", help="Flag to enable training via self-play (as opposed to fixed opponent)")
    parser.add_argument("--cpu-players", type=int, default=2, choices=[1, 2], help="Number of CPU players (options: 1, 2)")
    parser.add_argument("--player", type=int, default=0, choices=[0,1], help="Choose which player to play as: red = 0, yellow = 1")
    parser.add_argument("--record", action="store_true", help="Flag to save a recording of the game (game.gif)")
    parser.add_argument(
        "--win-rate",
        type=float,
        default=0.6,
        help="the expected winning rate: Optimal policy can get 0.7",
    )
    parser.add_argument(
        "--watch",
        default=False,
        action="store_true",
        help="no training, " "watch the play of pre-trained models",
    )
    parser.add_argument(
        "--agent-id",
        type=int,
        default=2,
        help="the learned agent plays as the"
        " agent_id-th player. Choices are 1 and 2.",
    )
    parser.add_argument(
        "--resume-path",
        type=str,
        default="",
        help="the path of agent pth file " "for resuming from a pre-trained agent",
    )
    parser.add_argument(
        "--opponent-path",
        type=str,
        default="",
        help="the path of opponent agent pth file "
        "for resuming from a pre-trained agent",
    )
    parser.add_argument(
        "--device", type=str, default="cuda" if torch.cuda.is_available() else "cpu"
    )
    return parser


def get_args() -> argparse.Namespace:
    parser = get_parser()
    return parser.parse_known_args()[0]


def get_agents(
    args: argparse.Namespace = get_args(),
    agent_learn: Optional[BasePolicy] = None,
    agent_opponent: Optional[BasePolicy] = None,
    optim: Optional[torch.optim.Optimizer] = None,
) -> Tuple[BasePolicy, torch.optim.Optimizer, list]:
    env = get_env()
    observation_space = (
        env.observation_space["observation"]
        if isinstance(env.observation_space, gym.spaces.Dict)
        else env.observation_space
    )
    args.state_shape = (
        observation_space["observation"].shape or observation_space["observation"].n
    )
    args.action_shape = env.action_space.shape or env.action_space.n
    if agent_learn is None:
        # model
        net = Net(
            args.state_shape,
            args.action_shape,
            hidden_sizes=args.hidden_sizes,
            device=args.device,
        ).to(args.device)
        if optim is None:
            optim = torch.optim.Adam(net.parameters(), lr=args.lr)
        agent_learn = DQNPolicy(
            net,
            optim,
            args.gamma,
            args.n_step,
            target_update_freq=args.target_update_freq,
        )
        if args.resume_path:
            agent_learn.load_state_dict(torch.load(args.resume_path))

    if agent_opponent is None:
        if args.self_play:
            # Create a new network with the same shape
            net_opponent = Net(
                args.state_shape,
                args.action_shape,
                hidden_sizes=args.hidden_sizes,
                device=args.device,
            ).to(args.device)
            agent_opponent = DQNPolicy(
                net_opponent,
                optim,
                args.gamma,
                args.n_step,
                target_update_freq=args.target_update_freq,
            )
        elif args.opponent_path:
            agent_opponent = deepcopy(agent_learn)
            agent_opponent.load_state_dict(torch.load(args.opponent_path))
        else:
            # agent_opponent = RandomPolicy()
            agent_opponent = GreedyPolicy() # Greedy policy is a difficult opponent, should yeild much better results than random

    if args.agent_id == 1:
        agents = [agent_learn, agent_opponent]
    else:
        agents = [agent_opponent, agent_learn]
    policy = MultiAgentPolicyManager(agents, env)
    return policy, optim, env.agents


def get_env(render_mode=None, args=None):
    return PettingZooEnv(gobblet_v1.env(render_mode=render_mode, args=args))


def train_agent(
    args: argparse.Namespace = get_args(),
    agent_learn: Optional[BasePolicy] = None,
    agent_opponent: Optional[BasePolicy] = None,
    optim: Optional[torch.optim.Optimizer] = None,
) -> Tuple[dict, BasePolicy]:
    # ======== environment setup =========
    train_envs = DummyVectorEnv([get_env for _ in range(args.training_num)])
    test_envs = DummyVectorEnv([get_env for _ in range(args.test_num)])
    # seed
    np.random.seed(args.seed)
    torch.manual_seed(args.seed)
    train_envs.seed(args.seed)
    test_envs.seed(args.seed)

    # ======== agent setup =========
    policy, optim, agents = get_agents(
        args, agent_learn=agent_learn, agent_opponent=agent_opponent, optim=optim
    )

    # ======== collector setup =========
    train_collector = Collector(
        policy,
        train_envs,
        VectorReplayBuffer(args.buffer_size, len(train_envs)),
        exploration_noise=True,
    )
    test_collector = Collector(policy, test_envs, exploration_noise=True)
    # policy.set_eps(1)
    train_collector.collect(n_step=args.batch_size * args.training_num)

    # ======== tensorboard logging setup =========
    log_path = os.path.join(args.logdir, "gobblet", "dqn")
    writer = SummaryWriter(log_path)
    writer.add_text("args", str(args))
    logger = TensorboardLogger(writer)

    # ======== callback functions used during training =========
    def save_best_fn(policy):
        if hasattr(args, "model_save_path"):
            model_save_path = args.model_save_path
        else:
            model_save_path = os.path.join(
                args.logdir, "gobblet", "dqn", "policy.pth"
            )
        torch.save(
            policy.policies[agents[args.agent_id - 1]].state_dict(), model_save_path
        )

    def stop_fn(mean_rewards):
        return mean_rewards >= args.win_rate

    def train_fn(epoch, env_step):
        policy.policies[agents[args.agent_id - 1]].set_eps(args.eps_train)

    def train_fn_selfplay(epoch, env_step):
        policy.policies[agents[0]].set_eps(args.eps_train) # Same as train_fn but for both agents instead of only learner
        policy.policies[agents[1]].set_eps(args.eps_train)

    def test_fn(epoch, env_step):
        policy.policies[agents[args.agent_id - 1]].set_eps(args.eps_test)

    def test_fn_selfplay(epoch, env_step):
        policy.policies[agents[0]].set_eps(args.eps_test) # Same as test_fn but for both agents instead of only learner
        policy.policies[agents[1]].set_eps(args.eps_test)


    def reward_metric(rews):
        return rews[:, args.agent_id - 1]

    # trainer
    result = offpolicy_trainer(
        policy,
        train_collector,
        test_collector,
        args.epoch,
        args.step_per_epoch,
        args.step_per_collect,
        args.test_num,
        args.batch_size,
        train_fn=train_fn_selfplay if args.self_play else train_fn,
        test_fn=test_fn_selfplay if args.self_play else test_fn,
        stop_fn=stop_fn,
        save_best_fn=save_best_fn,
        update_per_step=args.update_per_step,
        logger=logger,
        test_in_train=False,
        reward_metric=reward_metric,
    )

    return result, policy.policies[agents[args.agent_id - 1]]


# ======== a test function that tests a pre-trained agent ======
def watch(
    args: argparse.Namespace = get_args(),
    agent_learn: Optional[BasePolicy] = None,
    agent_opponent: Optional[BasePolicy] = None,
) -> None:
    env = DummyVectorEnv([lambda: get_env(render_mode=args.render_mode, args=args)])
    policy, optim, agents = get_agents(
        args, agent_learn=agent_learn, agent_opponent=agent_opponent
    )
    policy.eval()
    if not args.self_play:
        policy.policies[agents[args.agent_id - 1]].set_eps(args.eps_test)
    else:
        policy.policies[agents[0]].set_eps(args.eps_test)
        policy.policies[agents[1]].set_eps(args.eps_test)

    collector = Collector(policy, env, exploration_noise=True)
    pettingzoo_env = env.workers[0].env.env  # DummyVectorEnv -> Tianshou PettingZoo Wrapper -> PettingZoo Env
    if args.record:
        recorder = GIFRecorder()
        recorder.capture_frame(pettingzoo_env.unwrapped.screen)
    else:
        recorder = None

    while pettingzoo_env.agents:
        result = collector.collect(n_step=1, render=args.render)
        if recorder is not None:
            recorder.capture_frame(pettingzoo_env.unwrapped.screen)

        time.sleep(0.25)

        if collector.data.terminated or collector.data.truncated:
            rews, lens = result["rews"], result["lens"]
            print(f"Final reward: {rews[:, 0].mean()}, length: {lens.mean()} [{policy.policies[agents[0]]}]")
            print(f"Final reward: {rews[:, 1].mean()}, length: {lens.mean()} [{policy.policies[agents[1]]}]")
            if recorder is not None:
                recorder.end_recording(pettingzoo_env.unwrapped.screen)
                recorder = None

# ======== allows the user to input moves and play vs a pre-trained agent ======
def play(
        args: argparse.Namespace = get_args(),
        agent_learn: Optional[BasePolicy] = None,
        agent_opponent: Optional[BasePolicy] = None,
) -> None:
    env = DummyVectorEnv([lambda: get_env(render_mode=args.render_mode, args=args)])
    # env = get_env(render_mode=args.render_mode, args=args) # Throws error because collector looks for length, could just override though since I'm using my own collector
    policy, optim, agents = get_agents(
        args, agent_learn=agent_learn, agent_opponent=agent_opponent
    )
    policy.eval()
    policy.policies[agents[args.agent_id - 1]].set_eps(args.eps_test)

    # Experimental: let the CPU agent to continue training (TODO: check if this actually changes things meaningfully)
    # policy.policies[agents[args.agent_id - 1]].set_eps(args.eps_train)

    collector = ManualPolicyCollector(policy, env, exploration_noise=True) # Collector for CPU actions

    pettingzoo_env = env.workers[0].env.env # DummyVectorEnv -> Tianshou PettingZoo Wrapper -> PettingZoo Env

    if args.record:
        recorder = GIFRecorder()
    else:
        recorder = None
    manual_policy = gobblet_v1.ManualPolicy(env=pettingzoo_env, agent_id=args.player, recorder=recorder) # Gobblet keyboard input requires access to raw_env (uses functions from board)
    while pettingzoo_env.agents:
        agent_id = collector.data.obs.agent_id
        # If it is the players turn and there are less than 2 CPU players (at least one human player)
        if agent_id == pettingzoo_env.agents[args.player]:
            observation = {"observation": collector.data.obs.obs.flatten(),
                            "action_mask": collector.data.obs.mask.flatten()} # PettingZoo expects a dict with this format
            action = manual_policy(observation, agent_id)

            result = collector.collect_result(action=action.reshape(1), render=args.render)
        else:
            result = collector.collect(n_step=1, render=args.render)

            if recorder is not None:
                recorder.capture_frame(pettingzoo_env.unwrapped.screen)

        if collector.data.terminated or collector.data.truncated:
            rews, lens = result["rews"], result["lens"]
            print(f"Final reward: {rews[:, args.player].mean()}, length: {lens.mean()} [Human]")
            print(f"Final reward: {rews[:, 1-args.player].mean()}, length: {lens.mean()} [{policy.policies[agents[1-args.player]]}]")
            if recorder is not None:
                recorder.end_recording(pettingzoo_env.unwrapped.screen)
                recorder = None

if __name__ == "__main__":
    # train the agent and watch its performance in a match!
    args = get_args()
    print("Training agent...")
    result, agent = train_agent(args)
    print("Starting game...")
    if args.cpu_players == 2:
        watch(args, agent)
    else:
        play(args, agent)


# adapted from https://github.com/Farama-Foundation/PettingZoo/blob/master/tutorials/Tianshou/3_cli_and_logging.py
"""
This is a full example of using Tianshou with MARL to train agents, complete with argument parsing (CLI) and logging.

Author: Will (https://github.com/WillDudley)

Python version used: 3.8.10

Requirements:
pettingzoo == 1.22.0
git+https://github.com/thu-ml/tianshou
"""

import argparse
from typing import Optional, Tuple


import torch
from tianshou.env import DummyVectorEnv
from tianshou.env.pettingzoo_env import PettingZooEnv
from tianshou.policy import BasePolicy, DQNPolicy, MultiAgentPolicyManager, RandomPolicy


from gobblet import gobblet_v1
from gobblet.game.collector_manual_policy import ManualPolicyCollector
from gobblet.game.utils import GIFRecorder
from gobblet.game.greedy_policy import GreedyPolicy
import time


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument("--seed", type=int, default=1626)
    parser.add_argument("--eps-test", type=float, default=0.05)
    parser.add_argument("--eps-train", type=float, default=0.1)
    parser.add_argument("--buffer-size", type=int, default=20000)
    parser.add_argument("--lr", type=float, default=1e-4) # TODO: Changing this to 1e-5 for some reason makes it pause after 3 or 4 epochs
    parser.add_argument(
        "--gamma", type=float, default=0.9, help="a smaller gamma favors earlier win"
    )
    parser.add_argument("--n-step", type=int, default=3)
    parser.add_argument("--target-update-freq", type=int, default=320)
    parser.add_argument("--epoch", type=int, default=50)
    parser.add_argument("--step-per-epoch", type=int, default=1000)
    parser.add_argument("--step-per-collect", type=int, default=10)
    parser.add_argument("--update-per-step", type=float, default=0.1)
    parser.add_argument("--batch-size", type=int, default=64)
    parser.add_argument(
        "--hidden-sizes", type=int, nargs="*", default=[128, 128, 128, 128]
    )
    parser.add_argument("--training-num", type=int, default=10)
    parser.add_argument("--test-num", type=int, default=10)
    parser.add_argument("--logdir", type=str, default="log")
    parser.add_argument("--render", type=float, default=0.1)
    parser.add_argument("--render_mode", type=str, default="human", choices=["human","rgb_array", "text", "text_full"], help="Choose the rendering mode for the game.")
    parser.add_argument("--debug", action="store_true", help="Flag to enable to print extra debugging info")
    parser.add_argument("--self_play", action="store_true", help="Flag to enable training via self-play (as opposed to fixed opponent)")
    parser.add_argument("--cpu-players", type=int, default=1, choices=[1, 2], help="Number of CPU players (options: 1, 2)")
    parser.add_argument("--player", type=int, default=0, choices=[0,1], help="Choose which player to play as: red = 0, yellow = 1")
    parser.add_argument("--record", action="store_true", help="Flag to save a recording of the game (game.gif)")
    parser.add_argument("--depth", type=int, default=2, choices=[1, 2, 3], help="Search depth for greedy agent. Options: 1,2,3")
    parser.add_argument(
        "--win-rate",
        type=float,
        default=0.6,
        help="the expected winning rate: Optimal policy can get 0.7",
    )
    parser.add_argument(
        "--watch",
        default=False,
        action="store_true",
        help="no training, " "watch the play of pre-trained models",
    )
    parser.add_argument(
        "--agent-id",
        type=int,
        default=2,
        help="the learned agent plays as the"
        " agent_id-th player. Choices are 1 and 2.",
    )
    parser.add_argument(
        "--resume-path",
        type=str,
        default="",
        help="the path of agent pth file " "for resuming from a pre-trained agent",
    )
    parser.add_argument(
        "--opponent-path",
        type=str,
        default="",
        help="the path of opponent agent pth file "
        "for resuming from a pre-trained agent",
    )
    parser.add_argument(
        "--device", type=str, default="cuda" if torch.cuda.is_available() else "cpu"
    )
    return parser


def get_args() -> argparse.Namespace:
    parser = get_parser()
    return parser.parse_known_args()[0]


def get_agents() -> Tuple[BasePolicy, list]:
    env = get_env()
    agents = [GreedyPolicy(depth=args.depth), GreedyPolicy(depth=args.depth)]
    policy = MultiAgentPolicyManager(agents, env)
    return policy, env.agents


def get_env(render_mode=None, args=None):
    return PettingZooEnv(gobblet_v1.env(render_mode=render_mode, args=args))


# ======== watch two greedy agents play each other (tends to get stuck in loops) ======
def watch() -> None:
    env = DummyVectorEnv([lambda: get_env(render_mode=args.render_mode, args=args)])
    policy, agents = get_agents()

    collector = ManualPolicyCollector(policy, env)

    pettingzoo_env = env.workers[0].env.env # DummyVectorEnv -> Tianshou PettingZoo Wrapper -> PettingZoo Env
    if args.record:
        recorder = GIFRecorder()
    else:
        recorder = None

    while pettingzoo_env.agents:
        result = collector.collect(n_step=1, render=args.render)
        if recorder is not None:
            recorder.capture_frame(screen)
            time.sleep(0.25)

        if collector.data.terminated or collector.data.truncated:
            rews, lens = result["rews"], result["lens"]
            print(f"Final reward: {rews[:, 0].mean()}, length: {lens.mean()} [{policy.policies[agents[0]]}]")
            print(f"Final reward: {rews[:, 1].mean()}, length: {lens.mean()} [{policy.policies[agents[1]]}]")
            if recorder is not None:
                recorder.end_recording(screen)
                recorder = None

# ======== allows the user to input moves and play vs a greedy agent ======
def play() -> None:
    env = DummyVectorEnv([lambda: get_env(render_mode=args.render_mode, args=args)])

    policy, agents = get_agents()
    collector = ManualPolicyCollector(policy, env, exploration_noise=True) # Collector for CPU actions

    pettingzoo_env = env.workers[0].env.env # DummyVectorEnv -> Tianshou PettingZoo Wrapper -> PettingZoo Env
    if args.record:
        recorder = GIFRecorder()
    else:
        recorder = None
    manual_policy = gobblet_v1.ManualPolicy(env=pettingzoo_env, agent_id=args.player, recorder=recorder) # Gobblet keyboard input requires access to raw_env (uses functions from board)

    while pettingzoo_env.agents:
        agent_id = collector.data.obs.agent_id
        # If it is the players turn and there are less than 2 CPU players (at least one human player)
        if agent_id == pettingzoo_env.agents[args.player]:
            observation = {"observation": collector.data.obs.obs.flatten(),
                            "action_mask": collector.data.obs.mask.flatten()} # PettingZoo expects a dict with this format
            action = manual_policy(observation, agent_id)

            result = collector.collect_result(action=action.reshape(1), render=args.render)
        else:
            result = collector.collect(n_step=1, render=args.render)
            if recorder is not None:
                recorder.capture_frame(pettingzoo_env.unwrapped.screen)

        if collector.data.terminated or collector.data.truncated:
            rews, lens = result["rews"], result["lens"]
            print(f"\nFinal reward: {rews[:, args.player].mean()}, length: {lens.mean()} [Human]")
            print(f"Final reward: {rews[:, 1-args.player].mean()}, length: {lens.mean()} [{policy.policies[agents[1-args.player]]}]")
            if recorder is not None:
                recorder.end_recording(pettingzoo_env.unwrapped.screen)

if __name__ == "__main__":
    # train the agent and watch its performance in a match!
    args = get_args()
    print("Starting game...")
    if args.cpu_players == 2:
        watch()
    else:
        play()


import numpy as np

class Board:
    def __init__(self, squares=None):
        # internally self.board.squares holds a representation of the gobblet board, consisting of three stacked 3x3 boards, one for each piece size.
        # We flatten it for simplicity, from [3,3,3] to [27,]

        # An empty board with the first small piece in the top left corner and the second large piece in the bottom right corner would be:
        # [[1, 0, 0, 0, 0, 0, 0, 0, 0],
        #  [0, 0, 0, 0, 0, 0, 0, 0, 0],
        #  [0, 0, 0, 0, 0, 0, 0, 0, 6]]
        # In each of the three levels (small, medum, and large pieces), an empty 3x3 board is [0, 0, 0, 0, 0, 0, 0, 0, 0]
        # where indexes are column wise order
        # 0 3 6
        # 1 4 7
        # 2 5 8

        # Reshaped to three dimensions, this would look like:
        # [[[1, 0, 0],
        #   [0, 0, 0],
        #   [0, 0, 0]],
        #  [[0, 0, 0],
        #   [0, 0, 0],
        #   [0, 0, 0]],
        #  [[0, 0, 0],
        #   [0, 0, 0],
        #   [0, 0, 6]



        # empty -- 0
        # player 0 -- 1
        # player 1 -- -1 # Default: 2
        self.squares = np.zeros(27)
        self.squares_preview = np.zeros(27)

        # precommute possible winning combinations
        self.calculate_winners()

    def setup(self):
        self.calculate_winners()

    def get_action_from_pos_piece(self, pos, piece):
        if pos in range(9) and piece in range(1,7):
            return 9 * (piece - 1) + pos
        else:
            return -1

    # Return an action if an agent can place of the specified size in the specified location
    # Checks both possible pieces of that size, and returns -1 if neither can be placed (i.e., they are both covered)
    def get_action(self, pos, piece_size, agent_index):
        piece1 = piece_size * 2 - 1
        piece2 = piece_size * 2
        action1 = pos + 9 * (piece1 - 1)
        action2 = pos + 9 * (piece2 - 1)
        if self.is_legal(action1, agent_index):
            return action1
        elif self.is_legal(action2, agent_index):
            return action2
        else:
            return -1

    # To determine the position from an action, we take the number modulo 9, resulting in a number 0-8
    def get_pos_from_action(self, action):
        return action % 9

    # To determine the piece from an action i, we use floor division by 9 (i // 9), resulting in a number 0-5, where 1-2 represent small pieces, 3-4 represent medium pieces, and 5-6 represent large pieces.
    def get_piece_from_action(self, action):
        return (action // 9) + 1

    # To determine the size of a given piece p (1-6), we use floor division by 2 (p + 1 // 2) resulting in a number 1-3
    def get_piece_size_from_action(self, action):
        piece = self.get_piece_from_action(action)
        return (piece + 1) // 2

    # Returns the index on the board [0-26] for a given action
    def get_index_from_action(self, action):
        pos = self.get_pos_from_action(action)  # [0-8]
        piece_size = self.get_piece_size_from_action(action) # [1-3]
        return pos + 9 * (piece_size - 1) # [0-26]


    # Return true if an action is legal, false otherwise.
    def is_legal(self, action, agent_index=0):
        pos = self.get_pos_from_action(action) # [0-8]
        piece = self.get_piece_from_action(action) # [1-6]
        piece_size = self.get_piece_size_from_action(action) # [1-3]
        agent_multiplier = 1 if agent_index == 0 else -1

        board = self.squares.reshape(3, 9)
        # Check if this piece has been placed (if the piece number occurs anywhere on the level of that piece size)
        if any(board[piece_size-1] == piece * agent_multiplier):
            current_loc = np.where(board[piece_size-1] == piece * agent_multiplier)[0] # Returns array of values where piece is placed
            if len(current_loc) > 1:
                raise Exception("PIECE HAS BEEN USED TWICE")
                return False # Piece has been used twice (not valid)
            else:
                current_loc = current_loc[0] # Current location [0-27]
            # If this piece is currently covered, moving it is not a legal action

            if self.check_covered().reshape(3,9)[piece_size - 1][current_loc] == 1:
                return False

        # If this piece has not been placed
        # Check if the spot on the flat 3x3 board is open (we can definitely place in that case)
        flatboard = self.get_flatboard()
        if flatboard[pos] == 0:
            return True
        else:
            existing_piece_number = flatboard[pos] # [1-6]
            existing_piece_size = (abs(existing_piece_number) + 1) // 2 # [1-3]
            if piece_size > existing_piece_size:
                return True
            else:
                return False

    # Update the board with an agent's move
    def play_turn(self, agent_index, action):
        piece = self.get_piece_from_action(action)
        if agent_index == 1:
            piece = piece * -1
        index = self.get_index_from_action(action)

        # First: check if a move is legal or not
        if not self.is_legal(action, agent_index):
            return
        # If piece has already been placed, clear previous location
        if piece in self.squares:
            old_index = np.where(self.squares == piece)[0][0]
            self.squares[old_index] = 0
        self.squares[index] = piece
        return

    # Expects flat [9,] length array, from tic-tac-toe code
    def calculate_winners(self):
        winning_combinations = []
        indices = [x for x in range(0, 9)]

        # Vertical combinations
        winning_combinations += [
            tuple(indices[i : (i + 3)]) for i in range(0, len(indices), 3)
        ]

        # Horizontal combinations
        winning_combinations += [
            tuple(indices[x] for x in range(y, len(indices), 3)) for y in range(0, 3)
        ]

        # Diagonal combinations
        winning_combinations.append(tuple(x for x in range(0, len(indices), 4)))
        winning_combinations.append(tuple(x for x in range(2, len(indices) - 1, 2)))

        self.winning_combinations = winning_combinations

    def print(self):
        print(self.get_flatboard().reshape(3,3).transpose())

    # returns flattened board consisting of only top pieces (excluding pieces which are gobbled by other pieces)
    def get_flatboard(self):
        flatboard = np.zeros(9)
        board = self.squares.reshape(3, 9)
        for i in range(9):  # For every square in the 3x3 grid, find the topmost element (largest piece)
            top_piece_size = (np.amax(
                abs(board[:, i])))  # [-3, 2, 0] denotes a large piece gobbling a medium piece, this will return 3
            top_piece_index = list(abs(board[:, i])).index(
                top_piece_size)  # Get the row of the top piece (have to index [0]
            top_piece_color = np.sign(board[
                                          top_piece_index, i])  # Get the color of the top piece: 1 for player_1 and -1 for player_2, 0 for neither
            flatboard[i] = top_piece_color * top_piece_size # Simplify the board into only the top elements
        return flatboard

    # returns:
    # 0 for no winner
    # 1 -- agent 1 wins
    # -1 -- agent 2 wins
    def check_for_winner(self):
        winner = 0
        flatboard = self.get_flatboard()
        for combination in self.winning_combinations:
            states = []
            for index in combination:
                states.append(flatboard[index])
            if all(x > 0 for x in states):
                winner = 1
            if all(x < 0 for x in states):
                winner = -1
        return winner

    def check_game_over(self):
        winner = self.check_for_winner()
        if winner in [1, -1]:
            return True
        else:
            return False

    def check_covered(self): # Return a 27 length array indicating which positions have a piece which is covered
        board = self.squares.reshape(3, 9)
        covered = np.zeros((3, 9))
        for i in range(9): # Check small pieces
            if board[0, i] != 0 and (board[1, i] != 0 or board[2, i] != 0): # If there is a small piece, and either a large or medium piece covering it
                covered[0, i] = 1
        for i in range(9): # Check medium pieces
            if board[1, i] != 0 and board[2, i] != 0: # If there is a meidum piece and a large piece covering it
                covered[1, i] = 1
        covered[2, :] = 0 # Large pieces can't be covered
        # Doesn't matter about what color is covering it, you can't move that piece this turn (allows self-gobbling)
        return covered.flatten()

# DEBUG
    def print_pieces(self):
        open_indices = [i for i in range(len(self.squares)) if self.squares[i] == 0]
        open_squares = [np.where(self.get_flatboard() == 0)[0]]
        occupied_squares = [i % 9 for i in range(len(self.squares)) if self.squares[i] != 0] # List with entries 0-9
        movable_squares = [i % 9 for i in occupied_squares if self.check_covered()[i] == 0] # List with entries 0-9
        covered_squares = [i % 9 for i in np.where(self.check_covered() == 1)[0] ] # List with entries 0-9
        print("open_indices: ", open_indices)
        print("open_squares: ", open_squares)
        print("squares with pieces: ", occupied_squares)
        print("squares with uncovered pieces: ", movable_squares)
        print("squares with covered pieces: ", covered_squares)

    def __str__(self):
        return str(self.squares.reshape(3,3,3))



from typing import Any, Dict, Optional, Union
from copy import deepcopy
import numpy as np
import torch

from tianshou.data import Batch, ReplayBuffer, to_numpy, to_torch_as
from tianshou.policy import BasePolicy

from gobblet.game.board import Board

class GreedyPolicy(BasePolicy):
    """
    Basic greedy policy which checks if a move results in a victory, and if it sets the opponent up to win (or lose) in the next turn
    """

    def __init__(
        self,
        depth: Optional[int] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.board = None
        self.depth = depth

    def forward(
        self,
        batch: Batch,
        state: Optional[Union[dict, Batch, np.ndarray]] = None,
        input: str = "obs",
        **kwargs: Any,
    ) -> Batch:
        """Compute action over the given batch data.

        If you need to mask the action, please add a "mask" into batch.obs, for
        example, if we have an environment that has "0/1/2" three actions:
        ::

            batch == Batch(
                obs=Batch(
                    obs="original obs, with batch_size=1 for demonstration",
                    mask=np.array([[False, True, False]]),
                    # action 1 is available
                    # action 0 and 2 are unavailable
                ),
                ...
            )

        :return: A :class:`~tianshou.data.Batch` with "act" key, containing
            the greedy action.

        .. seealso::

            Please refer to :meth:`~tianshou.policy.BasePolicy.forward` for
            more detailed explanation.
        """
        full_batch = deepcopy(batch)
        if len(batch) > 1:
            acts = []
            for b in range(len(batch)):
                batch_single = deepcopy(full_batch)
                if len(batch_single[input].agent_id) > 1:
                    batch_single[input].obs = batch_single[input].obs[b, ...]
                    batch_single[input].mask = batch_single[input].mask[b, ...]
                    batch_single[input].agent_id = batch_single[input].agent_id[b, ...].reshape(1)
                act = self.forward_unbatched(batch=batch_single)  # array(1)
                acts.append(act) # [ array(1), array(2), ... array(10) ]
            acts = np.array(acts)
            return Batch(act=acts)  # Batch( act: [ array(1), array(2), ... array(10) ] )
        else:
            batch_single = deepcopy(full_batch)
            act = self.forward_unbatched(batch=batch_single)
            acts = np.array(act).reshape(1)
            return Batch(act=acts)

    def forward_unbatched(
        self,
        batch: Batch,
        input: str = "obs",
        **kwargs: Any,
    ) -> Batch:
        obs = batch[input]
        obs_depth2 = obs.obs if hasattr(obs, "obs") else obs  #

        obs_player = obs_depth2[..., :6]  # Index the last dimension
        board_player = np.array([(i + 1) * obs_player[..., i] + (i + 2) * obs_player[..., i + 1] for i in
                                 range(0, 6, 2)])  # Reshapes [3,3,6] to [3,3,3]
        obs_opponent = obs_depth2[..., 6:12]
        board_opponent = np.array(
            [(i + 1) * obs_opponent[..., i] + (i + 2) * obs_opponent[..., i + 1] for i in range(0, 6, 2)])
        board = np.where(board_player > board_opponent, board_player, -board_opponent)

        agents = ["player_1", "player_2"]
        agent_index = agents.index(obs.agent_id[0])
        opponent_index = 1 - agent_index  # If agent index is 1, we want 0; if agent index is 0, we want 1

        # If we are playing as the second agent, we need to adjust the representation of the board to reflect that
        # TODO: it would be cleaner to add a thirteenth channel that explicitly encodes the player whose turn it is
        # avoids issues of the wrong agent index (easier debugging) and enforces turn order in a stricter way
        if agent_index == 1:
            board = -board

        self.board = Board()
        self.board.squares = board.flatten().copy()

        # Map agent index to winner value that the board will return (agent idx: [0,1], winner_vals: [1,-1])
        winner_values = [1, -1]

        legal_actions = obs.mask.flatten().nonzero()[0]
        actions_depth1 = list(legal_actions) # Initialize the same as legal actions, then remove ones that cause a loss
        chosen_action = None

        results = {}
        # Depth 1
        for action in legal_actions:
            if self.board.is_legal(agent_index=agent_index, action=action):
                depth1_board = Board()
                depth1_board.squares = self.board.squares.copy()
                depth1_board.play_turn(agent_index=agent_index, action=action)

                # If we can win next turn, do it
                results[action] = depth1_board.check_for_winner()
                if results[action] == winner_values[agent_index]:  # Win for our agent
                    chosen_action = action
                    break
                elif results[action] == winner_values[opponent_index]:  # Loss for our agent
                    if len(actions_depth1) > 1:
                        actions_depth1.remove(action)
                    else:
                        break  # If there is nothing we can do to prevent them from winning, we have to do one of the moves

        if self.depth > 1:
            # Depth 2 (only explore cases where the immediate action doesn't result in a win or loss)
            for action in [ key for key, value in results.items() if value == 0 ]:
                # if self.board.is_legal(agent_index=agent_index, action=action):
                depth1_board = Board()
                depth1_board.squares = self.board.squares.copy()
                depth1_board.play_turn(agent_index=agent_index, action=action)

                # Check what the opponent can do after this potential move
                legal_actions_depth2 = [act for act in range(len(obs.mask.flatten())) if
                                      depth1_board.is_legal(agent_index=opponent_index, action=act)]

                results_depth2 = {}
                for action_depth2 in legal_actions_depth2:
                    depth2_board = Board()
                    depth2_board.squares = depth1_board.squares.copy()
                    depth2_board.play_turn(agent_index=opponent_index, action=action_depth2)

                    results_depth2[action_depth2] = depth2_board.check_for_winner()

                    # If the opponent can win in the next move, we don't want to do this action
                    if results_depth2[action_depth2] == winner_values[
                        opponent_index]:  # Check if it's possible for the opponent to win next turn after
                        if len(actions_depth1) > 1:
                            if action in actions_depth1:
                                actions_depth1.remove(action)
                        else:
                            break  # If there is nothing we can do to prevent them from winning, we have to pick one

                        # If we can put our piece in the place that the opponent would have gone to win the game, do that
                        # BUT this might miss us from winning the game ourselves in blcoking them, so don't exit the loop
                        # So we only do it if there aren't deterministic wins already chosen (and we don't break, keep looking after)
                        if self.board.is_legal(action_depth2, agent_index=agent_index):
                            if chosen_action == None:
                                chosen_action = action_depth2

                # Depth 2: if this move sets the opponent up with only moves that win the game for us, pick this move
                if all(winner == winner_values[agent_index] for winner in results_depth2.values()):
                    chosen_action = action
                    break
                # Depth 2: if this move blocks the opponent from winning, pick this if we cannot find any guaranteed wins
                if all(winner != winner_values[opponent_index] for winner in results_depth2.values()):
                    chosen_action = action # BLOCKING ACTION

                    # Depth 3: Given that we block the opponent this way, can we win the next turn?
                    if self.depth == 3:
                        depth1_board = Board()
                        depth1_board.squares = self.board.squares.copy()
                        depth1_board.play_turn(agent_index=agent_index, action=action)

                        # Search over depth 2 actions where the opponent doesn't win
                        for action_depth2 in [ key for key, value in results_depth2.items() if value == 0 ]:
                            depth2_board = Board()
                            depth2_board.squares = depth1_board.squares.copy()
                            depth2_board.play_turn(agent_index=agent_index, action=action_depth2)

                            legal_actions_depth3 = [act for act in range(len(obs.mask.flatten())) if
                                                depth2_board.is_legal(agent_index=agent_index, action=act)]
                            actions_depth3 = list(legal_actions_depth3)
                            results_depth3 = {}

                            # Search over all depth 3 actions which we can do
                            for action_depth3 in legal_actions_depth3:
                                depth3_board = Board()
                                depth3_board.squares = depth2_board.squares.copy()
                                depth3_board.play_turn(agent_index=agent_index, action=action)

                                # If we can win next turn, do it
                                results_depth3[action_depth3] = depth3_board.check_for_winner()
                                if results_depth3[action_depth3] == winner_values[agent_index]:  # Win for our agent
                                    chosen_action = action # If we can win in depth 3, then we know this blocking action is good
                                    break
                                elif results_depth3[action_depth3] == winner_values[opponent_index]:  # Loss for our agent
                                    if len(actions_depth3) > 1:
                                        actions_depth3.remove(action)
                                    else:
                                        break  # If there is nothing we can do to prevent them from winning in depth3, we have to do one of the moves

        if chosen_action is None:
            # Choose randomly between possible actions:
            chosen_action = np.random.choice(actions_depth1)
            # print(f"Choosing randomly between possible actions: {actions_depth1} --> {chosen_action}")
        act = np.array(chosen_action)
        return act

    def learn(self, batch: Batch, **kwargs: Any) -> Dict[str, float]:
        """Since a random agent learns nothing, it returns an empty dict."""
        return {}


# Modified from tutorial code
"""Basic code which shows what it's like to run PPO on the Pistonball env using the parallel API, this code is inspired by CleanRL.

This code is exceedingly basic, with no logging or weights saving.
The intention was for users to have a (relatively clean) ~200 line file to refer to when they want to design their own learning algorithm.

Author: Jet (https://github.com/jjshoots)
"""

import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
from torch.distributions.categorical import Categorical

from gobblet import gobblet_v1


class Agent(nn.Module):
    def __init__(self, num_actions):
        super().__init__()

        self.network = nn.Sequential(
            self._layer_init(nn.Conv2d(4, 32, 3, padding=1)),
            nn.MaxPool2d(2),
            nn.ReLU(),
            self._layer_init(nn.Conv2d(32, 64, 3, padding=1)),
            nn.MaxPool2d(2),
            nn.ReLU(),
            self._layer_init(nn.Conv2d(64, 128, 3, padding=1)),
            nn.MaxPool2d(2),
            nn.ReLU(),
            nn.Flatten(),
            self._layer_init(nn.Linear(128 * 8 * 8, 512)),
            nn.ReLU(),
        )
        self.actor = self._layer_init(nn.Linear(512, num_actions), std=0.01)
        self.critic = self._layer_init(nn.Linear(512, 1))

    def _layer_init(self, layer, std=np.sqrt(2), bias_const=0.0):
        torch.nn.init.orthogonal_(layer.weight, std)
        torch.nn.init.constant_(layer.bias, bias_const)
        return layer

    def get_value(self, x):
        return self.critic(self.network(x / 255.0))

    def get_action_and_value(self, x, action_mask=None, action=None):
        hidden = self.network(x / 255.0)
        logits = self.actor(hidden)
        print(logits.shape)
        print(len(action_mask))
        print(action_mask.shape)
        probs = Categorical(logits=logits)
        if action is None:
            action = probs.sample()
        return action, probs.log_prob(action), probs.entropy(), self.critic(hidden)


def batchify_obs(obs, device):
    """Converts PZ style observations to batch of torch arrays."""
    # convert to list of np arrays
    obs = np.stack([obs[a] for a in obs["observation"]], axis=0)
    # transpose to be (batch, channel, height, width)
    obs = obs.transpose(0, -1, 1, 2)
    # convert to torch
    obs = torch.tensor(obs).to(device)

    return obs

def batchify_action_mask(obs, device):
    """Converts PZ style observations to batch of torch arrays."""
    # convert to list of np arrays
    obs = np.stack([obs[a] for a in obs["action_mask"]], axis=0)
    # transpose to be (batch, channel, height, width)
    obs = obs.transpose(0, -1, 1, 2)
    # convert to torch
    obs = torch.tensor(obs).to(device)


def batchify(x, device):
    """Converts PZ style returns to batch of torch arrays."""
    # convert to list of np arrays
    x = np.stack([x[a] for a in x], axis=0)
    # convert to torch
    x = torch.tensor(x).to(device)

    return x


def unbatchify(x, env):
    """Converts np array to PZ style arguments."""
    x = x.cpu().numpy()
    x = {a: x[i] for i, a in enumerate(env.possible_agents)}

    return x


if __name__ == "__main__":

    """ALGO PARAMS"""
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    ent_coef = 0.1
    vf_coef = 0.1
    clip_coef = 0.1
    gamma = 0.99
    batch_size = 32
    stack_size = 4
    frame_size = (64, 64)
    max_cycles = 125
    total_episodes = 2

    """ ENV SETUP """
    env = gobblet_v1.parallel_env(render_mode="")
    num_agents = len(env.possible_agents)
    num_actions = env.action_space(env.possible_agents[0]).n
    observation_size = env.observation_space(env.possible_agents[0]).shape

    """ LEARNER SETUP """
    agent = Agent(num_actions=num_actions).to(device)
    optimizer = optim.Adam(agent.parameters(), lr=0.001, eps=1e-5)

    """ ALGO LOGIC: EPISODE STORAGE"""
    end_step = 0
    total_episodic_return = 0
    rb_obs = torch.zeros((max_cycles, num_agents, stack_size, *frame_size)).to(device)
    rb_actions = torch.zeros((max_cycles, num_agents)).to(device)
    rb_logprobs = torch.zeros((max_cycles, num_agents)).to(device)
    rb_rewards = torch.zeros((max_cycles, num_agents)).to(device)
    rb_terms = torch.zeros((max_cycles, num_agents)).to(device)
    rb_values = torch.zeros((max_cycles, num_agents)).to(device)

    """ TRAINING LOGIC """
    # train for n number of episodes
    for episode in range(total_episodes):

        # collect an episode
        with torch.no_grad():

            # collect observations and convert to batch of torch tensors
            env.reset(seed=None)
            next_obs, reward, termination, truncation, info = env.last() # Initial observation/info
            # reset the episodic return
            total_episodic_return = 0

            # each episode has num_steps
            for step in range(0, max_cycles):

                # rollover the observation
                obs = batchify_obs(next_obs, device)
                action_mask = batchify_action_mask(next_obs, device)

                # get action from the agent
                actions, logprobs, _, values = agent.get_action_and_value(obs, action_mask)

                # execute the environment and log data
                next_obs, rewards, terms, truncs, infos = env.step(
                    unbatchify(actions, env)
                )

                # add to episode storage
                rb_obs[step] = obs
                rb_rewards[step] = batchify(rewards, device)
                rb_terms[step] = batchify(terms, device)
                rb_actions[step] = actions
                rb_logprobs[step] = logprobs
                rb_values[step] = values.flatten()

                # compute episodic return
                total_episodic_return += rb_rewards[step].cpu().numpy()

                # if we reach termination or truncation, end
                if any([terms[a] for a in terms]) or any([truncs[a] for a in truncs]):
                    end_step = step
                    break

        # bootstrap value if not done
        with torch.no_grad():
            rb_advantages = torch.zeros_like(rb_rewards).to(device)
            for t in reversed(range(end_step)):
                delta = (
                    rb_rewards[t]
                    + gamma * rb_values[t + 1] * rb_terms[t + 1]
                    - rb_values[t]
                )
                rb_advantages[t] = delta + gamma * gamma * rb_advantages[t + 1]
            rb_returns = rb_advantages + rb_values

        # convert our episodes to batch of individual transitions
        b_obs = torch.flatten(rb_obs[:end_step], start_dim=0, end_dim=1)
        b_logprobs = torch.flatten(rb_logprobs[:end_step], start_dim=0, end_dim=1)
        b_actions = torch.flatten(rb_actions[:end_step], start_dim=0, end_dim=1)
        b_returns = torch.flatten(rb_returns[:end_step], start_dim=0, end_dim=1)
        b_values = torch.flatten(rb_values[:end_step], start_dim=0, end_dim=1)
        b_advantages = torch.flatten(rb_advantages[:end_step], start_dim=0, end_dim=1)

        # Optimizing the policy and value network
        b_index = np.arange(len(b_obs))
        clip_fracs = []
        for repeat in range(3):
            # shuffle the indices we use to access the data
            np.random.shuffle(b_index)
            for start in range(0, len(b_obs), batch_size):
                # select the indices we want to train on
                end = start + batch_size
                batch_index = b_index[start:end]

                _, newlogprob, entropy, value = agent.get_action_and_value(
                    b_obs[batch_index], b_actions.long()[batch_index]
                )
                logratio = newlogprob - b_logprobs[batch_index]
                ratio = logratio.exp()

                with torch.no_grad():
                    # calculate approx_kl http://joschu.net/blog/kl-approx.html
                    old_approx_kl = (-logratio).mean()
                    approx_kl = ((ratio - 1) - logratio).mean()
                    clip_fracs += [
                        ((ratio - 1.0).abs() > clip_coef).float().mean().item()
                    ]

                # normalize advantaegs
                advantages = b_advantages[batch_index]
                advantages = (advantages - advantages.mean()) / (
                    advantages.std() + 1e-8
                )

                # Policy loss
                pg_loss1 = -b_advantages[batch_index] * ratio
                pg_loss2 = -b_advantages[batch_index] * torch.clamp(
                    ratio, 1 - clip_coef, 1 + clip_coef
                )
                pg_loss = torch.max(pg_loss1, pg_loss2).mean()

                # Value loss
                value = value.flatten()
                v_loss_unclipped = (value - b_returns[batch_index]) ** 2
                v_clipped = b_values[batch_index] + torch.clamp(
                    value - b_values[batch_index],
                    -clip_coef,
                    clip_coef,
                )
                v_loss_clipped = (v_clipped - b_returns[batch_index]) ** 2
                v_loss_max = torch.max(v_loss_unclipped, v_loss_clipped)
                v_loss = 0.5 * v_loss_max.mean()

                entropy_loss = entropy.mean()
                loss = pg_loss - ent_coef * entropy_loss + v_loss * vf_coef

                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

        y_pred, y_true = b_values.cpu().numpy(), b_returns.cpu().numpy()
        var_y = np.var(y_true)
        explained_var = np.nan if var_y == 0 else 1 - np.var(y_true - y_pred) / var_y

        print(f"Training episode {episode}")
        print(f"Episodic Return: {np.mean(total_episodic_return)}")
        print(f"Episode Length: {end_step}")
        print("")
        print(f"Value Loss: {v_loss.item()}")
        print(f"Policy Loss: {pg_loss.item()}")
        print(f"Old Approx KL: {old_approx_kl.item()}")
        print(f"Approx KL: {approx_kl.item()}")
        print(f"Clip Fraction: {np.mean(clip_fracs)}")
        print(f"Explained Variance: {explained_var.item()}")
        print("\n-------------------------------------------\n")

    """ RENDER THE POLICY """
    env = gobblet_v1.parallel_env(render_mode="human", continuous=False)

    agent.eval()

    with torch.no_grad():
        # render 5 episodes out
        for episode in range(5):
            obs = batchify_obs(env.reset(seed=None), device)
            terms = [False]
            truncs = [False]
            while not any(terms) and not any(truncs):
                actions, logprobs, _, values = agent.get_action_and_value(obs)
                obs, rewards, terms, truncs, infos = env.step(unbatchify(actions, env))
                obs = batchify_obs(obs, device)
                terms = [terms[a] for a in terms]
                truncs = [truncs[a] for a in truncs]